# Resthys - Une interface REST partielle pour Noethys
# Copyright (C) 2018  Cliss XXI <tech@cliss21.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from django.core.checks import Error, Warning
from django.core.management import call_command
from django.conf import settings

import io

from difflib import unified_diff

no_models = Warning(
    "Unable to import resthys/base/models.py",
    hint='Verify your database settings and that your DB has been '
    'inspected.',
    obj='resthys.base',
    id='resthys.W003',
)


def check_noethys_db_version(app_configs, **kwargs):
    try:
        from resthys.base import models
    except Exception:
        return [no_models]
    try:
        db_version = models.Parametres.objects.get(nom='version').parametre
    except Exception:
        return [
            Warning(
                "Unable to get Noethys database version.",
                hint='Verify your database settings and that your DB is a '
                'Noethys database. '
                'After double checking, report an issue to <tech@cliss21.org>.',
                obj='resthys.base',
                id='resthys.W004',
            )
        ]
    if db_version not in settings.NOETHYS_VERSIONS:
        return [
            Warning(
                "Unsupported Noethys version",
                hint='Verify that your database version is well supported and '
                'update NOETHYS_VERSIONS in settings.',
                obj='resthys.base',
                id='resthys.W005',
            )
        ]
    return []


def check_inspected_database(app_configs, **kwargs):
    buf = io.StringIO()
    rc = call_command('inspectdb', '--database=noethys', stdout=buf)
    if rc is not None:
        return [
            Error(
                "Command inspectdb failed.",
                hint='Try to launch "./manage.py inspectdb --database=noethys" '
                'and report an issue to <tech@cliss21.org>.',
                obj='resthys.base',
                id='resthys.E001',
            )
        ]

    buf.seek(0)
    inspected_db = buf.read()
    buf.close()
    try:
        with open('resthys/base/models.py', 'r') as f:
            if f.read() != inspected_db:
                f.seek(0)
                for line in unified_diff(
                    f.read().split('\n'), inspected_db.split('\n')
                ):
                    print(line)
                return [
                    Warning(
                        "Outdated resthys/base/models.py",
                        hint='Refresh it using "make inspectdb" '
                        'and then run migrations.',
                        obj='resthys.base',
                        id='resthys.W002',
                    )
                ]
    except Exception:
        return [
            Warning(
                "Missing resthys/base/models.py",
                hint='Create it using "make inspectdb" '
                'and then run migrations.',
                obj='resthys.base',
                id='resthys.W001',
            )
        ]
    return []


def check_facture_template(app_configs, **kwargs):
    try:
        from resthys.base import models
    except Exception:
        return [no_models]
    objs = None
    try:
        d = models.DocumentsModeles.objects.get(categorie='facture', defaut=1)
        objs = models.DocumentsObjets.objects.filter(
            modele=d, categorie='bloc_texte', texte__contains='{FAMILLE_NOM}'
        )
    except Exception:
        pass

    if objs is None or len(objs) < 1:
        return [
            Warning(
                "Misconfigured factures",
                hint='It seems that {FAMILLE_NOM} is missing from default '
                'facture template configured in Noethys. Please verify it in '
                'Noethys via Parametres -> Modeles -> Modeles de documents',
                obj='resthys.base',
                id='resthys.E002',
            )
        ]
    return []
