# Resthys - Une interface REST partielle pour Noethys
# Copyright (C) 2018  Cliss XXI <tech@cliss21.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from rest_framework import serializers

from resthys.base import models, from_noethys


class UtilisateursSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Utilisateurs
        fields = [f.name for f in model._meta.fields]


class IndividusSerializer(serializers.ModelSerializer):
    sexe = serializers.SerializerMethodField()

    def get_sexe(self, obj):
        return from_noethys.get_sexe(obj)

    class Meta:
        model = models.Individus
        fields = ('id', 'nom', 'prenom', 'email', 'date_naiss', 'sexe')


class RattachementsSerializer(serializers.ModelSerializer):
    individu = IndividusSerializer()

    class Meta:
        model = models.Rattachements
        fields = ('id', 'individu', 'titulaire')


class FamillesSerializer(serializers.ModelSerializer):
    rattachements_set = RattachementsSerializer(many=True)
    titulaires_prettyprinted = serializers.SerializerMethodField()

    def get_titulaires_prettyprinted(self, obj):
        return from_noethys.get_titulaires_prettyprinted(obj)

    class Meta:
        model = models.Familles
        fields = ('id', 'rattachements_set', 'titulaires_prettyprinted')


class ActivitesSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Activites
        fields = ('id', 'nom')


class ReglementsSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Reglements
        fields = ('id', 'date', 'montant')


class VentilationSerializer(serializers.ModelSerializer):
    reglement = ReglementsSerializer()

    class Meta:
        model = models.Ventilation
        fields = [f.name for f in model._meta.fields]


class PrestationsSerializer(serializers.ModelSerializer):
    individu = IndividusSerializer()
    activite = ActivitesSerializer()
    montant = serializers.SerializerMethodField()
    regle = serializers.SerializerMethodField()
    solde = serializers.SerializerMethodField()

    def get_montant(self, obj):
        return from_noethys.get_prestation_montant(obj)

    def get_solde(self, obj):
        return from_noethys.get_prestation_solde(obj)

    def get_regle(self, obj):
        return from_noethys.get_prestation_regle(obj)

    class Meta:
        model = models.Prestations
        fields = (
            'id',
            'activite',
            'date',
            'individu',
            'label',
            'montant',
            'regle',
            'solde',
        )


class ComptesPayeursSerializer(serializers.ModelSerializer):
    individu = IndividusSerializer()
    famille = FamillesSerializer()
    adresse = serializers.SerializerMethodField()
    email = serializers.SerializerMethodField()

    def get_email(self, obj):
        return from_noethys.get_compte_payeur_email(obj)

    def get_adresse(self, obj):
        return from_noethys.get_compte_payeur_adresse(obj)

    class Meta:
        model = models.ComptesPayeurs
        fields = [f.name for f in model._meta.fields] + ['adresse', 'email']


class OrganisateurSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Organisateur
        fields = [
            f.name
            for f in model._meta.fields
            if f.name not in ['logo_update', 'logo']
        ]


class FacturesDetailSerializer(serializers.ModelSerializer):
    compte_payeur = ComptesPayeursSerializer()
    prestations_set = PrestationsSerializer(many=True)
    montant = serializers.SerializerMethodField()
    regle = serializers.SerializerMethodField()
    solde = serializers.SerializerMethodField()
    organisateur = serializers.SerializerMethodField()
    autres_factures = serializers.SerializerMethodField()

    def get_organisateur(self, obj):
        return OrganisateurSerializer(
            models.Organisateur.objects.get(pk=1)
        ).data

    def get_autres_factures(self, obj):
        return from_noethys.get_autres_factures(obj)

    def get_montant(self, obj):
        return from_noethys.get_facture_montant(obj)

    def get_solde(self, obj):
        return from_noethys.get_facture_solde(obj)

    def get_regle(self, obj):
        return from_noethys.get_facture_regle(obj)

    class Meta:
        model = models.Factures
        fields = (
            'id',
            'numero',
            'date_edition',
            'date_debut',
            'date_fin',
            'date_echeance',
            'compte_payeur',
            'prestations_set',
            'montant',
            'regle',
            'solde',
            'organisateur',
            'autres_factures',
        )


class FacturesListSerializer(serializers.ModelSerializer):
    montant = serializers.SerializerMethodField()

    def get_montant(self, obj):
        return from_noethys.get_facture_montant(obj)

    class Meta:
        model = models.Factures
        fields = (
            'id',
            'numero',
            'date_edition',
            'date_debut',
            'date_fin',
            'date_echeance',
            'montant',
        )


class FacturesPayementSerializer(serializers.ModelSerializer):
    SAISIE_CHOICES = (
        ('M', 'Formulaire Tipi'),
        ('T', 'Test'),
        ('X', 'Activation'),
        ('A', 'Compte Usager Tipi'),
    )

    RESULT_CHOICES = (('P', 'Payé'), ('R', 'Refusé'))

    solde = serializers.DecimalField(
        read_only=True, coerce_to_string=True, max_digits=7, decimal_places=2
    )

    amount = serializers.DecimalField(
        write_only=True, max_digits=7, decimal_places=2
    )
    email = serializers.EmailField(write_only=True, max_length=80, min_length=6)
    numcli = serializers.CharField(write_only=True, max_length=6, min_length=6)
    objet = serializers.CharField(write_only=True, max_length=99, min_length=6)
    refdet = serializers.CharField(write_only=True, max_length=30, min_length=6)
    saisie = serializers.ChoiceField(write_only=True, choices=SAISIE_CHOICES)
    result = serializers.ChoiceField(write_only=True, choices=RESULT_CHOICES)
    auto = serializers.CharField(write_only=True, max_length=6, min_length=6)
    related_invoices = serializers.CharField(write_only=True, max_length=99)
    date = serializers.DateTimeField(write_only=True)

    def update(self, request, *args, **kwargs):
        return False

    class Meta:
        model = models.Factures
        fields = (
            'solde',
            'amount',
            'email',
            'numcli',
            'objet',
            'refdet',
            'saisie',
            'result',
            'auto',
            'date',
            'related_invoices',
        )


# FIXME: temporaire
class FacturesUnpaySerializer(serializers.ModelSerializer):
    def update(self, request, *args, **kwargs):
        return False

    class Meta:
        model = models.Factures
        fields = ()
