# Resthys - Une interface REST partielle pour Noethys
# Copyright (C) 2018  Cliss XXI <tech@cliss21.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import inspect
import json
import uuid
from datetime import datetime

from django.test import TestCase, Client
from django.urls import reverse
from django.test.utils import override_settings
from django.contrib.auth.models import User
from rest_framework.authtoken.models import Token

from resthys.base import models

unique_name = "Boop6Eingiez"


class modelsLoadedTestCase(TestCase):
    def setUp(self):
        self.loaded_models = [
            m
            for _, m in models.__dict__.items()
            if inspect.isclass(m) and issubclass(m, models.models.Model)
        ]

    def test_primary_key_is_id(self):
        [self.assertEqual(m._meta.pk.name, 'id') for m in self.loaded_models]

    def test_foreign_key_do_not_startwith_id(self):
        # The inverse not has to be true: and id.* field might refers to
        # an hardcoded choice value
        for m in self.loaded_models:
            for r in m._meta.related_objects:
                self.assertFalse(r.name.startswith('id'))


class IndividusTestCase(TestCase):
    def setUp(self):
        self.tearDown()

        i = models.Individus.objects.create(nom=unique_name)
        a = models.Activites.objects.create(nom=unique_name)
        models.Consommations.objects.create(individu=i, activite=a)

    def test_created(self):
        i = models.Individus.objects.get(nom=unique_name)
        self.assertTrue(i.id)
        self.assertEqual(i.nom, unique_name)
        c = models.Consommations.objects.get(individu=i)
        self.assertTrue(c.id)
        self.assertEqual(c.activite.nom, unique_name)

    def test_individu_to_consommation_request(self):
        # get consommations from individu name
        r = models.Consommations.objects.filter(individu__nom=unique_name)
        self.assertEqual(len(r), 1)
        self.assertEqual(r[0].individu.nom, unique_name)
        self.assertEqual(r[0].activite.nom, unique_name)

    def test_activite_to_consommation_request(self):
        # get consommations from activite name
        r = models.Consommations.objects.filter(activite__nom=unique_name)
        self.assertEqual(len(r), 1)
        self.assertEqual(r[0].individu.nom, unique_name)
        self.assertEqual(r[0].activite.nom, unique_name)

    def test_activite_to_individu_request(self):
        # get all individu having a consommation with given activite name
        activites = models.Activites.objects.filter(nom=unique_name)
        consos = models.Consommations.objects.filter(activite__in=activites)
        r = models.Individus.objects.filter(consommations__in=consos)
        self.assertEqual(len(r), 1)
        self.assertEqual(r[0].nom, unique_name)

    def test_individu_deletion(self):
        i = models.Individus.objects.get(nom=unique_name)
        self.assertEqual(i.delete()[0], 1)
        self.assertEqual(
            len(models.Individus.objects.filter(nom=unique_name)), 0
        )

    def test_consommation_deletion(self):
        c = models.Consommations.objects.get(activite__nom=unique_name)
        self.assertEqual(c.delete()[0], 1)
        self.assertEqual(
            len(models.Consommations.objects.filter(activite__nom=unique_name)),
            0,
        )

    def test_activite_deletion(self):
        a = models.Activites.objects.get(nom=unique_name)
        self.assertEqual(a.delete()[0], 1)
        self.assertEqual(
            len(models.Activites.objects.filter(nom=unique_name)), 0
        )

    def tearDown(self):
        models.Activites.objects.filter(nom=unique_name).delete()
        models.Consommations.objects.filter(individu__nom=unique_name).delete()
        models.Individus.objects.filter(nom=unique_name).delete()


def get_or_create_user():
    try:
        return User.objects.get(username='resthys')
    except Exception:
        user = User.objects.create_user('resthys', None, 'resthyssy')
        return user


def get_or_create_token(user):
    try:
        token = Token.objects.get(user=user)
    except Exception:
        token = Token.objects.create(user=user)
    return token.key


class BasicAuthRestAPITestCase(TestCase):
    def setUp(self):
        get_or_create_user()
        self.client = Client()

    def test_deny_list_factures(self):
        response = self.client.get(reverse('factures_list'))
        self.assertEqual(response.status_code, 403)

    def test_access_list_factures(self):
        self.client.login(username='resthys', password='resthyssy')
        response = self.client.get(reverse('factures_list'))
        self.assertEqual(response.status_code, 200)


class TokenAuthRestAPITestCase(TestCase):
    def setUp(self):
        user = get_or_create_user()
        self.key = get_or_create_token(user)
        self.client = Client()

    def test_deny_list_factures(self):
        response = self.client.get(reverse('factures_list'))
        self.assertEqual(response.status_code, 403)

    def test_access_list_factures(self):
        response = self.client.get(
            reverse('factures_list'),
            HTTP_AUTHORIZATION='Token {}'.format(self.key),
        )
        self.assertEqual(response.status_code, 200)


class RestAPITestCase(TestCase):
    def setUp(self):
        self.tearDown()
        get_or_create_user()
        self.client = Client()
        self.client.login(username='resthys', password='resthyssy')

    def test_individu_email(self):
        url = reverse('individus_detail', args=['42'])
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        assert 'email' in response.json()
        response = self.client.put(
            url, json.dumps({'email': 'bling'}), content_type='application/json'
        )
        self.assertEqual(response.status_code, 400)
        response = self.client.put(
            url,
            json.dumps({'email': 'bling@bling.fr'}),
            content_type='application/json'
        )
        self.assertEqual(response.status_code, 200)

    def test_list_factures(self):
        tic = datetime.now()
        response = self.client.get(reverse('factures_list'))
        toc = datetime.now()
        self.assertEqual(response.status_code, 200)
        self.assertTrue((toc - tic).total_seconds() < 5)

    def test_list_some_factures(self):
        response = self.client.get(
            reverse('factures_list'), {'numero': ('000009', '000050', '000054')}
        )
        self.assertEqual(response.status_code, 200)
        r = response.json()
        self.assertEqual(len(r), 3)
        self.assertEqual(set([f['numero'] for f in r]), {9, 50, 54})

        response = self.client.get(
            reverse('factures_list'),
            {
                'numero': ('000009', '000050', '000054'),
                'nom': 'Isabelle Merignac',
            },
        )
        self.assertEqual(response.status_code, 200)
        r = response.json()
        self.assertEqual(len(r), 1)
        self.assertEqual(set([f['numero'] for f in r]), {54})

    def test_readonly_list_factures(self):
        url = reverse('factures_list')
        response = self.client.post(url)
        self.assertEqual(response.status_code, 405)
        response = self.client.put(url)
        self.assertEqual(response.status_code, 405)
        response = self.client.patch(url)
        self.assertEqual(response.status_code, 405)

    def test_facture_53_34(self):
        response = self.client.get(reverse('factures_detail', args=['000053']))
        self.assertEqual(response.status_code, 200)
        facture1 = response.json()
        self.assertEqual(facture1['numero'], 53)
        self.assertEqual(
            [{'numero': 34, 'solde': 14000}], facture1['autres_factures']
        )

        response = self.client.get(reverse('factures_detail', args=['000034']))
        self.assertEqual(response.status_code, 200)
        facture2 = response.json()
        self.assertEqual(facture2['numero'], 34)
        self.assertEqual(
            [{'numero': 53, 'solde': 3600}], facture2['autres_factures']
        )

        self.assertEqual(facture1['compte_payeur'], facture2['compte_payeur'])

    def test_facture_54(self):
        response = self.client.get(reverse('factures_detail', args=['000054']))
        self.assertEqual(response.status_code, 200)
        facture = response.json()
        self.assertEqual(facture['numero'], 54)
        self.assertEqual(
            facture['compte_payeur']['famille']['rattachements_set'][0][
                'individu'
            ]['nom'],
            "MERIGNAC",
        )
        for prestation in facture['prestations_set']:
            self.assertEqual(
                prestation['montant'], prestation['regle'] + prestation['solde']
            )
        montant = sum([p['montant'] for p in facture['prestations_set']])
        regle = sum([p['regle'] for p in facture['prestations_set']])
        solde = sum([p['solde'] for p in facture['prestations_set']])
        self.assertEqual(montant, regle + solde)
        self.assertEqual(montant, facture['montant'])
        self.assertEqual(regle, facture['regle'])
        self.assertEqual(solde, facture['solde'])
        self.assertEqual([], facture['autres_factures'])

    def test_readonly_facture_54(self):
        url = reverse('factures_detail', args=['000054'])
        response = self.client.post(url)
        self.assertEqual(response.status_code, 405)
        response = self.client.put(url)
        self.assertEqual(response.status_code, 405)
        response = self.client.patch(url)
        self.assertEqual(response.status_code, 405)

    def test_paiement_54(self):
        url = reverse('factures_paiement', args=['000054'])
        dummy_paiement_data = {
                'amount': '4200',
                'saisie': 'T',
                'related_invoices': '54',
                'objet': 'Facture bling',
                'refdet': str(uuid.uuid4())[0:30],
                'numcli': '000001',
                'email': 'bling@blang.fr',
                'date': '2018-09-11T11:14:59',
                'auto': '010203',
                }  # lack result to be complete
        response = self.client.put(url)
        self.assertEqual(response.status_code, 400)  # fails because bad arg
        response = self.client.put(
            url,
            json.dumps({**dummy_paiement_data}),
            content_type='application/json'
        )
        self.assertEqual(response.status_code, 400)  # fails because bad arg
        response = self.client.put(
            url,
            json.dumps({**dummy_paiement_data, 'result': 'X'}),
            content_type='application/json'
        )
        self.assertEqual(response.status_code, 400)  # fails because bad arg
        response = self.client.put(
            url,
            json.dumps({**dummy_paiement_data, 'result': 'P'}),
            content_type='application/json',
        )
        self.assertEqual(response.status_code, 200)
        response = self.client.get(reverse('factures_detail', args=['000054']))
        self.assertEqual(response.status_code, 200)
        facture = response.json()
        self.assertEqual(facture['solde'], 0)
        self.test_facture_54()

    def _test_paiement(self, numero):
        url = reverse('factures_paiement', args=[numero])
        dummy_paiement_data = {
                'amount': '4200',
                'saisie': 'T',
                'related_invoices': str(numero),
                'objet': 'Facture bling',
                'refdet': str(uuid.uuid4())[0:30],
                'numcli': '000001',
                'email': 'bling@blang.fr',
                'date': '2018-09-11T11:14:59',
                'auto': '010203',
                'result': 'P',
                }

        response = self.client.put(
            url,
            json.dumps(dummy_paiement_data),
            content_type='application/json'
        )
        self.assertEqual(response.status_code, 200)
        response = self.client.get(reverse('factures_detail', args=[numero]))
        self.assertEqual(response.status_code, 200)
        facture = response.json()
        self.assertEqual(facture['solde'], 0)
        self.assertEqual(facture['montant'], facture['regle'])

    def test_paiement_50(self):
        self._test_paiement('000050')

    def test_paiement_51(self):
        self._test_paiement('000051')

    def _test_unpaiement(self, numero):
        url = reverse('factures_unpay', args=[numero])
        response = self.client.put(
            url, json.dumps({}), content_type='application/json'
        )
        self.assertEqual(response.status_code, 200)
        response = self.client.get(reverse('factures_detail', args=[numero]))
        self.assertEqual(response.status_code, 200)
        facture = response.json()
        self.assertEqual(facture['regle'], 0)
        self.assertEqual(facture['montant'], facture['solde'])

    @override_settings(DEBUG=True)
    def test_unpaiement_50(self):
        self._test_unpaiement('000050')

    @override_settings(DEBUG=True)
    def test_unpaiement_51(self):
        self._test_unpaiement('000051')

    @override_settings(DEBUG=True)
    def test_reversion_52(self):
        numero = '000052'
        before = self.client.get(reverse('factures_detail', args=[numero]))
        url = reverse('factures_paiement', args=[numero])
        dummy_paiement_data = {
                'amount': '4200',
                'saisie': 'T',
                'related_invoices': str(numero),
                'objet': 'Facture bling',
                'refdet': str(uuid.uuid4())[0:30],
                'numcli': '000001',
                'email': 'bling@blang.fr',
                'date': '2018-09-11T11:14:59',
                'auto': '010203',
                'result': 'P',
                }

        self.client.put(
            url,
            json.dumps(dummy_paiement_data),
            content_type='application/json'
        )
        url = reverse('factures_unpay', args=[numero])
        self.client.put(
            url, json.dumps({}), content_type='application/json'
        )
        after = self.client.get(reverse('factures_detail', args=[numero]))
        self.assertEqual(before.content, after.content)

    def test_debug_needed_for_unpay(self):
        url = reverse('factures_unpay', args=['000054'])
        response = self.client.put(
            url, json.dumps({}), content_type='application/json'
        )
        self.assertEqual(response.status_code, 404)

    def test_named_facture_54(self):
        url = reverse('factures_detail', args=['000054'])
        for nom, expected in (
            ('Marc Mérignac', 200),
            ('Marc Merignac', 200),
            ('Marc  Merignac', 200),
            ('Mérignac Marc', 200),
            ('Mérignac Marc\'', 200),
            ('Mrignac Marc', 404),
            ('Marc', 404),
            ('M', 404),
            ('', 404),
            ('`OR 1=1;`Mérignac Marc', 404),
            ('Isabelle Mérignac', 200),
            ('Isabelle et Marc Mérignac', 200),
            ('Marc et Isabelle Mérignac', 200),
            ('Mérignac Isabelle et Marc', 200),
        ):
            response = self.client.get(url, {'nom': nom})
            self.assertEqual(response.status_code, expected)

    def tearDown(self):
        f = models.Factures.objects.get(numero=54)
        f.solde = 3300
        f.save()
        reglements = models.Reglements.objects.filter(
            observations='Créé par Resthys'
        )
        [[v.delete() for v in r.ventilation_set.all()] for r in reglements]
        reglements.delete()
        models.ModesReglements.objects.filter(label='TIPI par PIAF').delete()
        models.Utilisateurs.objects.filter(nom='Resthys').delete()
        models.Historique.objects.filter(action__contains='PIAF').delete()
