# Resthys - Une interface REST partielle pour Noethys
# Copyright (C) 2018  Cliss XXI <tech@cliss21.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import decimal

from resthys.base import models, from_noethys


"""
Noethys Payment

The main call for a entried payment is here:
    https://github.com/Noethys/Noethys/blob/master/noethys/Dlg/DLG_Saisie_reglement.py#L1469

It relies also on dispatching which is defined here:
    https://github.com/Noethys/Noethys/blob/master/noethys/Ctrl/CTRL_Ventilation.py#L774

When a paiement is saized, Noethys:
    - fetch attributes from UI (date, mode, emeteur, n° piece, montant, payeur,
      observation, compte, utilisateur)
    - fetch familly from invoice's compte_payeurs
    - compute fees
    - save the payment (reglement)
    - save dispatches (ventilation), i.e. for each prestation, add a
      corresponding ventilation
    - save fees
    - save history
"""


def create_or_get_tipi_mode(regie):
    label = 'Régie TIPI {}'.format(regie)
    r, _ = models.ModesReglements.objects.get_or_create(
        label=label, type_comptable='banque'
    )
    return r


def create_or_get_piaf_user():
    r, _ = models.Utilisateurs.objects.get_or_create(
        nom='PIAF', prenom='par CLISS XXI'
    )
    return r


def create_or_get_bank_account(regie):
    nom = 'DFT {}'.format(regie)
    r, _ = models.ComptesBancaires.objects.get_or_create(nom=nom)
    return r


def add_to_history(action, famille):
    return models.Historique.objects.create(
        heure=datetime.datetime.now().time(),
        date=datetime.date.today(),
        utilisateur=create_or_get_piaf_user(),
        famille=famille,
        categorie=None,
        action=action,
    )


def get_currency():
    try:
        return models.Parametres.objects.filter(parametre="monnaie_symbole")[0]
    except Exception:
        return '€'


def cents_to_euros(c):
    return decimal.Decimal(c).quantize(decimal.Decimal('0.01')).shift(-2)


def set_facture_paid(f, data):
    data['amount'] = cents_to_euros(data['amount'])

    r, _ = models.Reglements.objects.get_or_create(
        date=data['date'],
        date_saisie=datetime.date.today(),
        mode=create_or_get_tipi_mode(data['numcli']),
        numero_piece=data['refdet'],
        montant=data['amount'],
        compte_payeur=f.compte_payeur,
        payeur=f.compte_payeur.payeurs_set.all()[0]
        if f.compte_payeur.payeurs_set.all() else None,
        observations='{} (saisie:{}; notification: {})'.format(
            data['objet'], data['saisie'], data['email']
        ),
        compte=create_or_get_bank_account(data['numcli']),
        utilisateur=create_or_get_piaf_user(),
    )
    solde_ventile = cents_to_euros(sum([
        from_noethys.get_ventilation_montant(v)
        for v in models.Ventilation.objects.filter(reglement=r)
        ]))
    for p in models.Prestations.objects.filter(facture=f):
        solde = cents_to_euros(from_noethys.get_prestation_solde(p))
        montant = min(solde, data['amount'] - solde_ventile)
        if montant > 0 and solde > 0:
            models.Ventilation.objects.create(
                compte_payeur=f.compte_payeur,
                reglement=r,
                prestation=p,
                montant=montant,
            )
        solde_ventile += montant
    add_to_history(
        'Saisie du règlement ID{} : '
        '{} {} en CB - payé en ligne via PIAF'.format(
            r.id, get_currency(), r.montant
        ),
        f.compte_payeur.famille,
    )


def set_facture_unpaid(f):
    reglements = set()
    for p in models.Prestations.objects.filter(facture=f):
        for v in models.Ventilation.objects.filter(prestation=p):
            if v.reglement:
                reglements.add(v.reglement)
            v.delete()
    for r in reglements:
        add_to_history(
            'Annulation du règlement ID{} : {} - en ligne via PIAF'.format(
                r.id, get_currency()
            ),
            f.compte_payeur.famille,
        )
        r.delete()
